@@ -123,7 +123,7 @@ static gboolean g_timeout_check        (gpointer  source_data,
 					GTimeVal *current_time,
 					gpointer  user_data);
 static gboolean g_timeout_dispatch     (gpointer  source_data,
-					GTimeVal *current_time,
+					GTimeVal *dispatch_time,
 					gpointer  user_data);
 static gboolean g_idle_prepare         (gpointer  source_data, 
 					GTimeVal *current_time,
@@ -133,7 +133,7 @@ static gboolean g_idle_check           (gpointer  source_data,
 					GTimeVal *current_time,
 					gpointer  user_data);
 static gboolean g_idle_dispatch        (gpointer  source_data,
-					GTimeVal *current_time,
+					GTimeVal *dispatch_time,
 					gpointer  user_data);
 
 /* Data */
@@ -626,7 +626,7 @@ g_get_current_time (GTimeVal *result)
 
 /* HOLDS: main_loop_lock */
 static void
-g_main_dispatch (GTimeVal *current_time)
+g_main_dispatch (GTimeVal *dispatch_time)
 {
   while (pending_dispatches != NULL)
     {
@@ -654,7 +654,7 @@ g_main_dispatch (GTimeVal *current_time)
 
 	  G_UNLOCK (main_loop);
 	  need_destroy = ! dispatch (source_data,
-				     current_time,
+				     dispatch_time,
 				     hook_data);
 	  G_LOCK (main_loop);
 
@@ -704,7 +704,7 @@ g_main_iterate (gboolean block,
 		gboolean dispatch)
 {
   GHook *hook;
-  GTimeVal current_time  ={ 0, 0 };
+  GTimeVal current_time  = { 0, 0 };
   gint n_ready = 0;
   gint current_priority = 0;
   gint timeout;
@@ -723,7 +723,7 @@ g_main_iterate (gboolean block,
       G_UNLOCK (main_loop);
       return FALSE;
     }
-#endif G_THREADS_ENABLED  
+#endif
   
   /* If recursing, finish up current dispatch, before starting over */
   if (pending_dispatches)
@@ -807,6 +807,9 @@ g_main_iterate (gboolean block,
 
   g_main_poll (timeout, n_ready > 0, current_priority);
 
+  if (timeout != 0)
+    g_get_current_time (&current_time);
+  
   /* Check to see what sources need to be dispatched */
 
   n_ready = 0;
@@ -1230,6 +1233,22 @@ g_main_wakeup (void)
 
 /* Timeouts */
 
+static void
+g_timeout_set_expiration (GTimeoutData *data,
+			  GTimeVal     *current_time)
+{
+  guint seconds = data->interval / 1000;
+  guint msecs = data->interval - seconds * 1000;
+
+  data->expiration.tv_sec = current_time->tv_sec + seconds;
+  data->expiration.tv_usec = current_time->tv_usec + msecs * 1000;
+  if (data->expiration.tv_usec >= 1000000)
+    {
+      data->expiration.tv_usec -= 1000000;
+      data->expiration.tv_sec++;
+    }
+}
+
 static gboolean 
 g_timeout_prepare  (gpointer  source_data, 
 		    GTimeVal *current_time,
@@ -1242,9 +1261,21 @@ g_timeout_prepare  (gpointer  source_data,
   msec = (data->expiration.tv_sec  - current_time->tv_sec) * 1000 +
          (data->expiration.tv_usec - current_time->tv_usec) / 1000;
 
-  *timeout = (msec <= 0) ? 0 : msec;
+  if (msec < 0)
+    msec = 0;
+  else if (msec > data->interval)
+    {
+      /* The system time has been set backwards, so we 
+       * reset the expiration time to now + data->interval;
+       * this at least avoids hanging for long periods of time.
+       */
+      g_timeout_set_expiration (data, current_time);
+      msec = data->interval;
+    }
 
-  return (msec <= 0);
+  *timeout = msec;
+
+  return (msec == 0);
 }
 
 static gboolean 
@@ -1261,23 +1292,14 @@ g_timeout_check (gpointer  source_data,
 
 static gboolean
 g_timeout_dispatch (gpointer source_data, 
-		    GTimeVal *current_time,
+		    GTimeVal *dispatch_time,
 		    gpointer user_data)
 {
   GTimeoutData *data = source_data;
 
   if (data->callback (user_data))
     {
-      guint seconds = data->interval / 1000;
-      guint msecs = data->interval - seconds * 1000;
-
-      data->expiration.tv_sec = current_time->tv_sec + seconds;
-      data->expiration.tv_usec = current_time->tv_usec + msecs * 1000;
-      if (data->expiration.tv_usec >= 1000000)
-	{
-	  data->expiration.tv_usec -= 1000000;
-	  data->expiration.tv_sec++;
-	}
+      g_timeout_set_expiration (data, dispatch_time);
       return TRUE;
     }
   else
@@ -1291,24 +1313,14 @@ g_timeout_add_full (gint           priority,
 		    gpointer       data,
 		    GDestroyNotify notify)
 {
-  guint seconds;
-  guint msecs;
   GTimeoutData *timeout_data = g_new (GTimeoutData, 1);
+  GTimeVal current_time;
 
   timeout_data->interval = interval;
   timeout_data->callback = function;
-  g_get_current_time (&timeout_data->expiration);
+  g_get_current_time (&current_time);
 
-  seconds = timeout_data->interval / 1000;
-  msecs = timeout_data->interval - seconds * 1000;
-
-  timeout_data->expiration.tv_sec += seconds;
-  timeout_data->expiration.tv_usec += msecs * 1000;
-  if (timeout_data->expiration.tv_usec >= 1000000)
-    {
-      timeout_data->expiration.tv_usec -= 1000000;
-      timeout_data->expiration.tv_sec++;
-    }
+  g_timeout_set_expiration (timeout_data, &current_time);
 
   return g_source_add (priority, FALSE, &timeout_funcs, timeout_data, data, notify);
 }
@@ -1344,7 +1356,7 @@ g_idle_check    (gpointer  source_data,
 
 static gboolean
 g_idle_dispatch (gpointer source_data, 
-		 GTimeVal *current_time,
+		 GTimeVal *dispatch_time,
 		 gpointer user_data)
 {
   GSourceFunc func = source_data;
